import abc
import six
from collections import namedtuple
import numpy as np
from toon.input.clock import mono_clock
import inspect
import ctypes


@six.add_metaclass(abc.ABCMeta)
class Obs():
    @property
    @abc.abstractmethod
    def shape(self):
        return None

    @property
    @abc.abstractmethod
    def ctype(self):
        return None
    # single observation

    def __init__(self, time, data):
        try:
            self.time = float(time)  # what if time is not a double?
            # is reshape expensive? should we just trust they did it right?
            self.data = np.asarray(data, dtype=self.ctype)
            self.data.shape = self.shape  # will error if mismatch?
        except (TypeError, ValueError):  # time or data not right
            self.time = None
            self.data = None

    def __repr__(self):
        return 'type: %s\ntime: %f\ndata: %s\nshape: %s\nctype: %s' % (type(self).__name__, self.time, self.data, self.shape, self.ctype)

    def __str__(self):
        return '%s(time: %f, data: %s)' % (type(self).__name__, self.time, self.data)

    def any(self):
        return not (self.time is None or self.data is None)


@six.add_metaclass(abc.ABCMeta)
class BaseDevice():
    def __init__(self, clock=mono_clock.get_time, **kwargs):
        _obs = self.__class__.get_obs()
        self.Returns = BaseDevice.build_named_tuple(_obs)
        self.clock = clock

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def read(self):
        pass

    sampling_frequency = 500

    # helpers to figure out the data returned by device
    # (without instantiation--key b/c we need to do *before* we instantiate on other process)
    @classmethod
    def get_obs(cls):
        # get all user-defined Obs defined within the class (as long as they don't start w/ double underscore)
        # separated from tuple building so that
        return [getattr(cls, p) for p in dir(cls) if not p.startswith('__')
                and not p.startswith('_abc')
                and not isinstance(getattr(cls, p), property)
                and inspect.isclass(getattr(cls, p))
                and issubclass(getattr(cls, p), Obs)]

    @classmethod
    def build_named_tuple(cls, obs):
        if obs:
            class Returns(namedtuple('Returns', [x.__name__.lower() for x in obs])):
                def any(self):
                    # simplify user checking of whether there's any data
                    return any([x is not None and x.time is not None for x in self])
            # default values of namedtuple to None (see mouse.py for example why)
            Returns.__new__.__defaults__ = (None,) * len(Returns._fields)
            return Returns
        return None



import ctypes
import gc
import multiprocessing as mp
import os
from collections import namedtuple
from copy import copy
from itertools import compress
from sys import platform
from toon.input.tsarray import TsArray as TsA

import numpy as np
import psutil

ctypes_map = np.ctypeslib._typecodes.copy()
ctypes_map['|S1'] = ctypes.c_char  # TODO: I have no clue if this is valid
ctypes_map['|b1'] = ctypes.c_bool


def shared_to_numpy(mp_arr, dims, dtype):
    """Convert a :class:`multiprocessing.Array` to a numpy array.
    Helper function to allow use of a :class:`multiprocessing.Array` as a numpy array.
    Derived from the answer at:
    <https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory-for-multiprocessing>
    """
    return np.frombuffer(mp_arr.get_obj(), dtype=dtype).reshape(dims)


class MpDevice(object):
    def __init__(self, device=None, high_priority=True,
                 buffer_len=None, **device_kwargs):
        """
        buffer_len overrides default (1000 samples) and sampling_frequency-based (1s worth)
        """
        self.device = device
        self.device_kwargs = device_kwargs
        self.buffer_len = buffer_len
        self.high_priority = high_priority

    def start(self):
        # For Macs, use spawn (interaction with OpenGL or ??)
        # Windows only does spawn
        if platform == 'darwin' or platform == 'win32':
            try:
                mp.set_start_method('spawn')
            except (AttributeError, RuntimeError):
                pass  # already started a process, or on python2

        n_buffers = 2
        self.shared_locks = []
        # make one lock per buffer
        for i in range(n_buffers):
            self.shared_locks.append(mp.RLock())
        self.remote_ready = mp.Event()  # signal to main process that remote is done setup
        self.kill_remote = mp.Event()  # signal to remote process to die
        self.remote_done = mp.Event()
        self.current_buffer_index = mp.Value(ctypes.c_bool, 0, lock=False)  # shouldn't need a lock

        # figure out number of observations to save between reads
        nrow = 100
        if self.device.sampling_frequency:
            nrow = self.device.sampling_frequency
        # if the user specified a runtime sampling_frequency, use that
        kwarg_freq = self.device_kwargs.get('sampling_frequency', None)
        if kwarg_freq:
            nrow = kwarg_freq
        if self.buffer_len:  # buffer_len overcomes all
            nrow = self.buffer_len
        nrow = max(int(nrow), 1)  # make sure we have at least one row
        _device_obs = self.device.get_obs()
        # use this tuple to return data
        self._return_tuple = self.device.build_named_tuple(_device_obs)
        self.nt = namedtuple('obs', 'time data')
        self._data = [[] for i in range(n_buffers)]  # one set of data per buffer
        # for each observation type, allocate arrays (though note that
        # there's nothing useful in them right now)
        for i in range(n_buffers):
            for obs in _device_obs:
                self._data[i].append(DataGlob(obs.ctype, obs.shape, nrow, self.shared_locks[i]))

        self._res = [None] * len(self._data[0])

        self.process = mp.Process(target=remote,
                                  args=(self.device, self.device_kwargs, self._data,
                                        self.remote_ready, self.kill_remote, os.getpid(),
                                        self.current_buffer_index, self.remote_done))
        self.process.daemon = True
        self.process.start()
        self.ps_process = psutil.Process(self.process.pid)
        self.original_nice = self.ps_process.nice()
        self.set_high_priority(self.high_priority)  # lame try to set priority
        self.remote_ready.wait()  # pause until the remote says it's ready
        for i in range(n_buffers):
            for obs in self._data[i]:
                obs.generate_squeeze()

    # @profile
    def read(self):
        # get the currently used buffer
        # note that the value only changes *if* this function has acquired the
        # lock, so we should always be safe to access w/o lock here
        current_buffer_index = int(self.current_buffer_index.value)
        # this *may* block, if the remote is currently writing
        with self._data[current_buffer_index][0].counter.get_lock():
            for counter, datum in enumerate(self._data[current_buffer_index]):
                datum.local_count = datum.counter.value
                datum.counter.value = 0  # reset (so that we start writing to top of array)
                if datum.local_count > 0:
                    np.copyto(datum.local_data.time, datum.np_data.time)
                    np.copyto(datum.local_data.T, datum.np_data.data.T)  # allow copying to 1D
                    self._res[counter] = datum.local_data[0:datum.local_count]
                else:
                    self._res[counter] = None
        # if only one Obs, don't stuff into a namedtuple
        if len(self._res) == 1:
            return self._res[0]
        # plug values into namedtuple
        return self._return_tuple(*self._res)

    def stop(self):
        self.set_high_priority(False)
        self.kill_remote.set()
        self.remote_done.wait()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()

    def set_high_priority(self, val=False):
        try:
            if val:
                if psutil.WINDOWS:
                    self.ps_process.nice(psutil.HIGH_PRIORITY_CLASS)
                else:
                    self.ps_process.nice(-10)
            else:
                self.ps_process.nice(self.original_nice)
        except (psutil.AccessDenied, psutil.NoSuchProcess) as e:
            pass


def remote(device, device_kwargs, shared_data,
           # extras
           remote_ready, kill_remote, parent_pid,  # don't include in coverage, implicitly tested
           current_buffer_index, remote_done):  # pragma: no cover

    def process_data():
        shared = shared_data[buffer_index][counter]  # alias
        if shared.counter.value < shared.nrow:  # haven't filled buffer yet, so next available row
            next_index = shared.counter.value
            shared.np_data.time[next_index] = datum.time
            shared.np_data.data[next_index, :] = datum.data
        else:  # rolling buffer, had cursed my bedroom
            np.copyto(shared.np_data.time, np.roll(shared.np_data.time, -1, axis=0))
            shared.np_data.time[-1] = datum.time
            np.copyto(shared.np_data.data, np.roll(shared.np_data.data, -1, axis=0))
            shared.np_data.data[-1, :] = datum.data
        # successful read, increment the indexing counter for this stream of data
        shared.counter.value += 1

    gc.disable()
    dev = device(**device_kwargs)

    with dev:
        remote_ready.set()  # signal to the local process that remote is ready to go
        while not kill_remote.is_set() and psutil.pid_exists(parent_pid):
            data = dev.read()  # get observation(s) from device
            buffer_index = int(current_buffer_index.value)  # can only change later
            if isinstance(data, list):  # if a list of observations, rather than a single one
                inds = [[d is not None and d.any() for d in l] for l in data]
                flag = any([any(d) for d in inds])
                is_list = True
            else:
                inds = [d is not None and d.any() for d in data]
                flag = any(inds)
                is_list = False
            if flag:  # any data at all (otherwise, don't bother acquiring locks)
                # test whether the current buffer is accessible
                lck = shared_data[buffer_index][0].counter.get_lock()
                success = lck.acquire(block=False)
                if not success:  # switch to the other buffer (the local process is in the midst of reading)
                    current_buffer_index.value = not current_buffer_index.value
                    buffer_index = int(current_buffer_index.value)
                    lck = shared_data[buffer_index][0].counter.get_lock()
                    lck.acquire()
                try:
                    if is_list:
                        for dat, ind in zip(data, inds):
                            for counter, datum in zip(np.where(ind)[0], compress(dat, ind)):
                                process_data()
                    else:
                        for counter, datum in zip(np.where(inds)[0], compress(data, inds)):
                            process_data()
                finally:
                    lck.release()
    remote_done.set()


# make sure this is visible for pickleability
obs = namedtuple('obs', 'time data')


class DataGlob(object):
    def __init__(self, ctype, shape, nrow, lock):
        self.new_dims = (nrow,) + shape
        self.ctype = ctypes_map[np.dtype(ctype).str]
        self.shape = shape
        self.is_scalar = self.shape == (1,)
        prod = int(np.prod(self.new_dims))
        # don't touch (usually)
        self.nrow = int(nrow)
        self._mp_data = obs(time=mp.Array(ctypes.c_double, self.nrow, lock=lock),
                            data=mp.Array(self.ctype, prod, lock=lock))
        self.counter = mp.Value(ctypes.c_uint, 0, lock=lock)
        self.generate_np_version()
        self.generate_local_version()

    def generate_np_version(self):
        self.np_data = obs(time=shared_to_numpy(self._mp_data.time, (self.nrow,), ctypes.c_double),
                           data=shared_to_numpy(self._mp_data.data, self.new_dims, self.ctype))

    def generate_local_version(self):
        self.local_count = 0
        self.local_data = TsA(self.np_data.data.copy(), time=self.np_data.time.copy())

    def generate_squeeze(self):
        # if scalar data, give the user a 1D array (rather than 2D)
        if self.is_scalar:
            self.local_data = TsA(np.squeeze(self.np_data.data.copy()), time=self.np_data.time.copy())



from collections import namedtuple
from toon.anim.easing import linear
from toon.anim.interpolators import lerp, select


class Track(object):
    def __init__(self, data, interpolator=lerp, easing=linear):
        # data is list of tuples
        self.data = data
        self.interpolator = interpolator
        self.easing = easing
        # if data is non-numeric, force user to use select
        if not isinstance(data[0][1], (float, int)):
            self.interpolator = select
            self.easing = linear
        self.state = 'stopped'

    def start(self, time):
        self.time = time
        self.state = 'playing'

    def update(self, time):
        if self.state == 'stopped':
            return

        # find the two keyframes to interpolate between
        # rel_time is the time relative to the start
        rel_time = time - self.time
        try:
            index, goal = next((i, x) for i, x in enumerate(self.data) if x[0] > rel_time)
        except StopIteration:
            self.state = 'stopped'
            return self.data[-1][1]
        # prev keyframe to check against
        reference = self.data[index - 1]
        # calculate time stuff
        goal_time = goal[0] - reference[0]
        new_time = rel_time - reference[0]
        time_warp = self.easing(1 - ((goal_time - new_time)/goal_time))
        return self.interpolator(reference[1], goal[1], time_warp)



from inspect import ismethod
from collections import namedtuple
from copy import copy

TrackAttr = namedtuple('TrackAttr', 'track attr obj kwargs')


class Player(object):
    def __init__(self, *args, **kwargs):
        self.tracks = {}

    def add(self, name, track, attr, obj=None, **kwargs):
        if name in self.tracks:
            raise ValueError('Track name already exists.')
        self.tracks.update({name: TrackAttr(copy(track), attr, obj, kwargs)})

    def remove(self, names):
        if isinstance(names, list):
            for i in names:
                self.tracks.pop(i, None)
        else:
            self.tracks.pop(names, None)
        return

    def start(self, time, names=None):
        if not names:
            for i in self.tracks:
                self.tracks[i].track.start(time)
        elif isinstance(names, list):
            for i in names:
                self.tracks[i].track.start(time)
        else:
            self.tracks[names].track.start(time)

    def _do_update(self, attr, val, obj, **kwargs):
        # if we get a function, call function with updated val
        if callable(attr):
            # if it's a method (e.g. a setter), we call that method with the val as an arg
            if ismethod(attr):
                if kwargs:
                    attr(val, **kwargs)
                else:
                    attr(val)
                return
            if kwargs:
                attr(val, obj, **kwargs)
            else:
                attr(val, obj)
            return
        # otherwise (user gave string), directly set the attribute
        setattr(obj, attr, val)

    def update(self, time):
        for i in self.tracks:
            # if tracks are playing, will return a val
            val = self.tracks[i].track.update(time)
            if val is not None:
                if self.tracks[i].obj:  # object or list provided, so we'll manipulate them
                    try:  # see if single object
                        self._do_update(self.tracks[i].attr, val,
                                        self.tracks[i].obj, **self.tracks[i].kwargs)
                    except (TypeError, AttributeError):  # list of objects?
                        for obj in self.tracks[i].obj:
                            print(obj)
                            self._do_update(self.tracks[i].attr, val,
                                            obj, **self.tracks[i].kwargs)
                else:  # operate on self
                    self._do_update(self.tracks[i].attr, val, self, **self.tracks[i].kwargs)

    def stop(self, names=None):
        if not names:
            for i in self.tracks:
                self.tracks[i].track.state = 'stopped'
        elif isinstance(names, list):
            for i in names:
                self.tracks[i].track.state = 'stopped'
        else:
            # single key
            self.tracks[names].track.state = 'stopped'

    def is_playing(self, name):
        return self.tracks[name].track.state == 'playing'

    def any_playing(self):
        return any([self.tracks[x].track.state == 'playing' for x in self.tracks])



from math import pow, sin, pi
# from https://github.com/mosra/magnum/blob/master/src/Magnum/Animation/Easing.h

pihalf = pi/2.0


def linear(x):
    return x


def step(x):
    return 0.0 if x < 0.5 else 1.0


def smoothstep(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    return (3.0 - 2.0 * x) * pow(x, 2)


def smootherstep(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    return pow(x, 3) * (x * (x * 6.0 - 15.0) + 10.0)


def quadratic_in(x):
    return pow(x, 2)


def quadratic_out(x):
    return -x * (x - 2.0)


def quadratic_in_out(x):
    if (x < 0.5):
        return 2.0*pow(x, 2)
    return 1.0 - 2.0*pow(1.0 - x, 2)


def exponential_in(x):
    return 0.0 if x <= 0.0 else pow(2.0, 10.0 * (x - 1.0))


def exponential_out(x):
    return 1.0 if x >= 1.0 else 1.0 - pow(2.0, -10.0 * x)


def exponential_in_out(x):
    if x <= 0.0:
        return 0.0
    if x >= 1.0:
        return 1.0
    if x < 0.5:
        return 0.5 * pow(2.0, 20.0*x - 10.0)
    return 1.0 - 0.5 * pow(2.0, 10.0 - 20.0 * x)


def elastic_in(x):
    return pow(2.0, 10.0 * (x - 1.0)) * sin(13.0 * pihalf * x)


def elastic_out(x):
    return 1.0 - pow(2.0, -10.0 * x) * sin(13.0 * pihalf * (x + 1.0))


def elastic_in_out(x):
    if x < 0.5:
        return 0.5 * pow(2.0, 10.0 * (2.0 * x - 1.0)) * sin(13.0 * pi * x)
    return 1.0 - 0.5 * pow(2.0, 10.0 * (1.0 - 2.0 * x)) * sin(13.0 * pi * x)



from time import sleep
import numpy as np
from toon.input.mpdevice import MpDevice
from tests.input.mockdevices import Dummy, DummyList, SingleResp

# bump up the sampling frequency for tests
Dummy.sampling_frequency = 1000
DummyList.sampling_frequency = 1000


def test_device_single():
    # single device with two data sources
    dev = MpDevice(Dummy)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)


def test_single_resp():
    dev = MpDevice(SingleResp)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(isinstance(res, np.ndarray))
    assert(type(res.time) is np.ndarray)


def test_device_list():
    # two observations per read on the device
    dev = MpDevice(DummyList)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)
    assert(res.num1.shape[0] > res.num2.shape[1])


def test_restart():
    # start & stop device
    dev = MpDevice(Dummy)
    dev.start()
    sleep(0.2)
    res = dev.read()
    dev.stop()
    dev.start()
    sleep(0.2)
    res2 = dev.read()
    dev.stop()
    assert(res.any() and res2.any())


def test_context():
    # device as context manager
    dev = MpDevice(Dummy)
    with dev:
        sleep(0.2)
        res = dev.read()
    assert(len(res.num1.time) > 5)
    assert(len(res.num1.time) == res.num1.shape[0])
    assert(res.num1.shape[1] == 5)
    assert(res.num2.dtype == np.int32)


def test_multi_devs():
    # 2+ devices at once (each gets own process)
    dev1 = MpDevice(Dummy)
    dev2 = MpDevice(Dummy)
    with dev1, dev2:
        sleep(0.1)
        res1 = dev1.read()
        res2 = dev2.read()

    assert(res1.num1.time is not None)
    assert(res2.num1.time is not None)
